"""
Database configuration and session creation.
"""
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from .config import get_settings

settings = get_settings()

if "sqlite" in settings.db_url:
    connect_args = {"check_same_thread": False}
else:
    connect_args = {}

engine = create_engine(
    settings.db_url,
    echo=settings.db_echo,
    pool_size=settings.db_pool_size,
    max_overflow=settings.db_max_overflow,
    pool_timeout=settings.db_pool_timeout,
    pool_recycle=settings.db_pool_recycle,
    pool_pre_ping=settings.db_pool_pre_ping,
    connect_args=connect_args,
)

SessionLocal = sessionmaker(bind=engine)

Base = declarative_base()


def get_db():
    """
    Returns a new session of the database.
    """
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()



"""
Configuration file for the users utils repository.
"""
from os import getenv
from functools import lru_cache
from pydantic import ConfigDict
from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    """
    App settings class.
    """
    model_config: ConfigDict = ConfigDict(
        env_file=".env" if not getenv("DB_URL") else None, extra="allow"
    )
    db_url: str = "postgresql+psycopg2:"\
        "//admin:admin@users_db_container/users_db"\
        if not getenv("DB_URL") else getenv("DB_URL")
    db_echo: bool = False
    db_pool_size: int = 5
    db_max_overflow: int = 10
    db_pool_timeout: int = 30
    db_pool_recycle: int = 3600
    db_pool_pre_ping: bool = True


@lru_cache
def get_settings() -> Settings:
    """
    Returns the app settings.
    """
    return Settings()



"""
Validator functions for th
"""
import re
from rut_chile import rut_chile


def validate_chilean_phone_number(number):
    """
    Validates a Chilean phone number.
    """
    if not re.match(r"^\+?56 \d{1} \d{4} \d{4}$", number):
        raise ValueError("phone must be in the format '+56 X XXXX XXXX'.")
    return number


def validate_rut(rut):
    """
    Validates a Chilean RUT.
    """
    if not re.match(r"^\d{1,2}\.\d{3}\.\d{3}-[\dkK]$", rut):
        raise ValueError("rut must be in the format 'XX.XXX.XXX-X'.")
    if not rut_chile.is_valid_rut(rut.upper()):
        raise ValueError("rut is not valid.")
    return rut



"""
Recurrent CRUD functions for apps that use the user model.
"""
from typing import Optional
from sqlalchemy.orm import Session

from . import models


def get_user_by_email_if_exists(
        db: Session, email: str) -> Optional[models.User]:
    """
    Returns a user by email if it exists.
    """
    return db.query(models.User).filter(models.User.email == email).first()


def get_admin_by_email_if_exists(
        db: Session, email: str) -> Optional[models.Admin]:
    """
    Returns an admin by email if it exists.
    """
    return db.query(models.Admin).filter(models.Admin.email == email).first()



"""
Schemas for the auth app
"""
import re
from enum import Enum
from typing import Optional
from datetime import date
from dateutil.relativedelta import relativedelta
from email_validator import validate_email, EmailNotValidError
from pydantic import BaseModel, Field, EmailStr, ConfigDict,\
    HttpUrl, field_serializer, field_validator


from .validators import validate_chilean_phone_number, validate_rut


PASS_REGEX_PATTERN = (
    r"^(?=.*[a-z])(?=.*[A-Z])(?=.*\d)"
    r"(?=.*[@$!%*#?&+\-_.;,])[A-Za-z\d@$!#%*?&+\-_.;,]{6,20}$"
)


class GenderEnum(str, Enum):
    """
    Enum for the gender attribute.
    """
    M = "Masculino"
    F = "Femenino"


class UserBase(BaseModel):
    """
    Base schema for the User model.
    """
    email: EmailStr
    name: str = Field(..., min_length=1, max_length=50)
    last_name: str = Field(..., min_length=1, max_length=50)
    phone: str = Field(..., min_length=12, max_length=16)
    rut: str = Field(..., min_length=11, max_length=12)
    birth_date: date
    gender: Optional[GenderEnum]
    region: Optional[str] = None
    city: Optional[str] = None
    commune: Optional[str] = None
    address: Optional[str] = None
    address_number: Optional[str] = None
    address_apartment_number: Optional[str] = None
    registered_since: Optional[date] = Field(..., default_factory=date.today)
    subscribed_since: Optional[date] = None
    is_subscribed: Optional[bool] = False
    profile_picture_url: Optional[HttpUrl] = \
        "https://chamaclub-bucket.s3.amazonaws.com/avatares/Astronout.webp"

    @field_validator('email')
    @classmethod
    def validate_email(cls, email: EmailStr):
        """
        Validates the email attribute.
        """
        try:
            validate_email(email)
        except EmailNotValidError as e:
            raise ValueError("Invalid email.") from e
        return email

    @field_validator('rut')
    @classmethod
    def validate_rut(cls, rut: str):
        """
        Validates the rut attribute.
        """
        validate_rut(rut)
        return rut

    @field_validator('phone')
    @classmethod
    def validate_phone(cls, phone: str):
        """
        Validates the phone attribute.
        """
        phone = phone.replace(" ", "")
        phone = phone[:3] + " " + phone[3:4] + " " +\
            phone[4:8] + " " + phone[8:]
        validate_chilean_phone_number(phone)
        return phone

    @field_serializer('profile_picture_url')
    def serialize_profile_picture_url(
            self, profile_picture_url: HttpUrl, _info):
        """
        Serializes the profile_picture_url attribute into a string.
        """
        if profile_picture_url is None:
            return None
        return str(profile_picture_url)

    @field_validator('birth_date')
    @classmethod
    def validate_birth_date(cls, birth_date: date):
        """
        Validates the birth_date attribute.
        """
        if birth_date > date.today() - relativedelta(years=18):
            raise ValueError("birth_date must be at least 18 years ago.")
        return birth_date


class UserCreate(UserBase):
    """
    Schema for creating a user.
    """
    password: str = Field(..., min_length=8, max_length=50)
    subscription_provider: str
    subscription_plan: str

    @field_validator('password')
    @classmethod
    def validate_password(cls, password: str):
        """
        Validates the password attribute.
        """
        if not re.match(PASS_REGEX_PATTERN, password):
            raise ValueError(
                "Password must contain at least one lowercase letter, "
                "one uppercase letter, one digit, and one special character."
            )
        return password


class User(UserBase):
    """
    Schema for the User model.
    """
    model_config: ConfigDict = ConfigDict(
        from_attributes=True
    )
    id: int = Field(..., gt=0)


class AdminBase(BaseModel):
    """
    Base schema for the Admin model.
    """
    email: EmailStr
    name: str = Field(..., min_length=1, max_length=50)


class AdminCreate(AdminBase):
    """
    Schema for creating an admin.
    """
    password: str = Field(..., min_length=8, max_length=50)

    @field_validator('password')
    @classmethod
    def validate_password(cls, password: str):
        """
        Validates the password attribute.
        """
        if not re.match(PASS_REGEX_PATTERN, password):
            raise ValueError(
                "Password must contain at least one lowercase letter, "
                "one uppercase letter, one digit, and one special character."
            )
        return password


class Admin(AdminBase):
    """
    Schema for the Admin model.
    """
    model_config: ConfigDict = ConfigDict(
        from_attributes=True
    )
    id: int = Field(..., gt=0)



"""
This module is the entry point for the user_utils package.
"""
from . import crud, database, models, schemas, validators

__all__ = [
    'crud',
    'database',
    'models',
    'schemas',
    'validators'
]



"""
Setup file for user_utils package
"""
from setuptools import setup, find_packages

setup(
    name='user_vpetrov',
    version='0.2.5',
    packages=find_packages(),
    install_requires=[
        'fastapi',
        'pydantic',
        'sqlalchemy',
        'alembic',
        'email-validator',
        'python-dateutil',
        'passlib',
        'rut-chile'
    ]
)



"""
User models.
"""
from datetime import date
from dateutil.relativedelta import relativedelta
from sqlalchemy import Column, Integer, Boolean, String, Date,\
    Enum, DateTime, ForeignKey
from sqlalchemy.orm import validates, relationship
from email_validator import validate_email, EmailNotValidError
from passlib.context import CryptContext

from .validators import validate_chilean_phone_number, validate_rut
from .schemas import GenderEnum
from .database import Base


pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")


class User(Base):
    """
    User model.
    """
    __tablename__ = "users"

    id = Column(Integer, primary_key=True, index=True)
    email = Column(String, unique=True, index=True, nullable=False)
    name = Column(String(50), nullable=False)
    last_name = Column(String(50), nullable=False)
    hashed_password = Column(String, nullable=False)
    phone = Column(String(15), nullable=False)
    rut = Column(String(12), unique=True, nullable=False)
    birth_date = Column(Date, nullable=False)
    gender = Column(Enum(GenderEnum), nullable=True, default=None)
    region = Column(String(50), nullable=True, default=None)
    city = Column(String(50), nullable=True, default=None)
    commune = Column(String(50), nullable=True, default=None)
    address = Column(String(100), nullable=True, default=None)
    address_number = Column(String(50), nullable=True, default=None)
    address_apartment_number = Column(String(50), nullable=True, default=None)
    registered_since = Column(Date, nullable=False)
    subscribed_since = Column(Date, nullable=True, default=None)
    is_subscribed = Column(Boolean, nullable=False, default=False)
    profile_picture_url = Column(String, nullable=True, default=None)

    @validates("name")
    def validate_name(self, _, name):
        """
        Validates the name attribute.
        """
        if not name:
            raise ValueError("name must not be empty.")
        return name

    @validates("last_name")
    def validate_last_name(self, _, last_name):
        """
        Validates the last_name attribute.
        """
        if not last_name:
            raise ValueError("last_name must not be empty.")
        return last_name

    @validates("email")
    def validate_email(self, _, email):
        """
        Validates the email attribute.
        """
        try:
            validate_email(email, check_deliverability=True)
        except EmailNotValidError as e:
            raise ValueError("email must not be empty.") from e
        return email

    @validates("phone")
    def validate_phone_is_chilean_phone(self, _, phone):
        """
        Validates the phone attribute.
        """
        validate_chilean_phone_number(phone)
        return phone

    @validates("rut")
    def validate_rut_input(self, _, rut):
        """
        Validates the rut attribute.
        """
        validate_rut(rut)
        return rut

    @validates("birth_date")
    def validate_birth_date(self, _, birth_date):
        """
        Validates the birth_date attribute.
        """
        if birth_date > date.today() - relativedelta(years=18):
            raise ValueError("birth_date must be at least 18 years ago.")
        return birth_date

    def verify_password(self, password: str) -> bool:
        """
        Verifies the user's password.
        """
        return pwd_context.verify(password, self.hashed_password)

    @classmethod
    def get_hashed_password(cls, password: str):
        """
        Returns the hashed password.
        """
        return pwd_context.hash(password)


class Admin(Base):
    """
    Admin model.
    """
    __tablename__ = "admins"

    id = Column(Integer, primary_key=True, index=True)
    email = Column(String, unique=True, index=True, nullable=False)
    name = Column(String(50), nullable=False)
    hashed_password = Column(String, nullable=False)

    two_factor = relationship(
        "TwoFactor", back_populates="admin", uselist=False,
        cascade="all, delete-orphan")

    @validates("email")
    def validate_email(self, _, email):
        """
        Validates the email attribute.
        """
        try:
            validate_email(email, check_deliverability=True)
        except EmailNotValidError as e:
            raise ValueError("email must not be empty.") from e
        return email

    @validates("name")
    def validate_name(self, _, name):
        """
        Validates the name attribute.
        """
        if not name:
            raise ValueError("name must not be empty.")
        return name

    def verify_password(self, password: str) -> bool:
        """
        Verifies the user's password.
        """
        return pwd_context.verify(password, self.hashed_password)

    @classmethod
    def get_hashed_password(cls, password: str):
        """
        Returns the hashed password.
        """
        return pwd_context.hash(password)


class TwoFactor(Base):
    """
    Two factor model.
    """
    __tablename__ = "two_factors"

    id = Column(Integer, primary_key=True, index=True)
    admin_id = Column(
        Integer, ForeignKey("admins.id"),
        nullable=False, unique=True, index=True)
    token = Column(String, nullable=False, unique=True)
    validation_code = Column(String(6), nullable=False)
    expiration_time = Column(DateTime, nullable=False)
    confirmed = Column(Boolean, nullable=False, default=False)

    admin = relationship(
        "Admin", back_populates="two_factor",
        uselist=False, foreign_keys=[admin_id])

    @validates("admin_id")
    def validate_admin_id(self, _, admin_id):
        """
        Validates the admin_id attribute.
        """
        if not admin_id:
            raise ValueError("admin_id must not be empty.")
        return admin_id

    @validates("token")
    def validate_token(self, _, token):
        """
        Validates the token attribute.
        """
        if not token:
            raise ValueError("token must not be empty.")
        return token
